#include <config.h>
#include <sys/types.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <assert.h>
#include <setjmp.h>
#include <pwd.h>
#include <getopt.h>
#include <signal.h>
#include <selinux/selinux.h>
#include <wchar.h>
# define SA_NOCLDSTOP 0
# define sigprocmask(How, Set, Oset) 
# define sigset_t int
#  define siginterrupt(sig, flag) 
# define SA_RESTART 0
#include "system.h"
#include <fnmatch.h>
#include "acl.h"
#include "argmatch.h"
#include "dev-ino.h"
#include "error.h"
#include "filenamecat.h"
#include "hard-locale.h"
#include "hash.h"
#include "human.h"
#include "filemode.h"
#include "filevercmp.h"
#include "idcache.h"
#include "ls.h"
#include "mbswidth.h"
#include "mpsort.h"
#include "obstack.h"
#include "quote.h"
#include "smack.h"
#include "stat-size.h"
#include "stat-time.h"
#include "strftime.h"
#include "xdectoint.h"
#include "xstrtol.h"
#include "areadlink.h"
#include "mbsalign.h"
#include "dircolors.h"
#define PROGRAM_NAME (ls_mode == LS_LS ? "ls" \
                      : (ls_mode == LS_MULTI_COL \
                         ? "dir" : "vdir"))
#define AUTHORS \
  proper_name ("Richard M. Stallman"), \
  proper_name ("David MacKenzie")
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#define longdiff(a, b) ((a) < (b) ? -1 : (a) > (b))
# define READDIR_LIES_ABOUT_MOUNTPOINT_D_INO 1
# define RELIABLE_D_INO(dp) D_INO (dp)
# define st_author st_uid
enum filetype
  {
    
    
    
    
    arg_directory
  };
#define FILETYPE_INDICATORS				\
  {							\
    C_ORPHAN, C_FIFO, C_CHR, C_DIR, C_BLK, C_FILE,	\
    C_LINK, C_SOCK, C_FILE, C_DIR			\
  }
enum acl_type
  {
    
    
    ACL_T_YES
  };
struct fileinfo
  {
    char *name;
    char *linkname;
    struct stat stat;
    enum filetype filetype;
    mode_t linkmode;
    char *scontext;
    bool stat_ok;
    bool linkok;
    enum acl_type acl_type;
    bool has_capability;
  };
#define LEN_STR_PAIR(s) sizeof (s) - 1, s
struct bin_str
  {
    size_t len;			
    const char *string;		
  };
# define tcgetpgrp(Fd) 0
static size_t quote_name (FILE *out, const char *name,
                          struct quoting_options const *options,
                          size_t *width);
static char *make_link_name (char const *name, char const *linkname);



static void put_indicator (const struct bin_str *ind);
static void attach (char *dest, const char *dirname, const char *name);


static void get_link_name (char const *filename, struct fileinfo *f,
                           bool command_line_arg);
static void indent (size_t from, size_t to);
static size_t calculate_columns (bool by_columns);
static size_t print_file_name_and_frills (const struct fileinfo *f,
                                          size_t start_col);
static void print_horizontal (void);


;
static void prep_non_filename_text (void);
;
#define INITIAL_TABLE_SIZE 30

#define LOOP_DETECT (!!active_dir_set)

static size_t  cwd_n_used, tabsize, line_length, max_idx;
static void **sorted_file;
static bool color_symlink_as_referent,  format_needs_type,  format_needs_stat,  qmark_funny_chars,  immediate_dirs,  print_inode,  check_symlink_color,  used_color = false,  print_with_color,  file_output_block_size = 1,  file_human_output_opts,  output_block_size,  human_output_opts,  print_block_size,  print_group = true,  print_author,  print_owner = true,  file_size_width,  minor_device_number_width,  major_device_number_width,  author_width,  group_width,  owner_width,  scontext_width,  nlink_width,  block_size_width,  inode_number_width,  any_has_acl,  print_scontext;
#define FILE_OR_LINK_MODE(File) \
    ((color_symlink_as_referent && (File)->linkok) \
     ? (File)->linkmode : (File)->stat.st_mode)
struct pending
  {
  };
static char UNKNOWN_SECURITY_CONTEXT[] = "?";
enum format
  {
    		
    		
    		
    			
    with_commas			
  };
static enum format format;
enum time_style
  {
    	
    	
    		
    locale_time_style		
  };
enum time_type
  {
    			
    time_atime,			
  };
enum sort_type
  {
    		
    		
    			
    		
    sort_time,			
  };
enum indicator_style
  {
    none,	
    	
    	
    	
  };
static enum indicator_style indicator_style;
enum color_type
  {
    		
    		
    color_if_tty		
  };
enum Dereference_symlink
  {
    	
    	
    DEREF_ALWAYS			
  };
enum indicator_no
  {
           
     
           
        C_MULTIHARDLINK,
  };
struct color_ext_type
  {
    struct bin_str ext;		
    struct bin_str seq;		
    struct color_ext_type *next;	
  };
static struct bin_str color_indicator[] =
  {
    { LEN_STR_PAIR ("\033[") },		
    { LEN_STR_PAIR ("m") },		
    { 0, NULL },			
    { LEN_STR_PAIR ("0") },		
    { 0, NULL },			
    { 0, NULL },			
    { LEN_STR_PAIR ("01;34") },		
    { LEN_STR_PAIR ("01;36") },		
    { LEN_STR_PAIR ("33") },		
    { LEN_STR_PAIR ("01;35") },		
    { LEN_STR_PAIR ("01;33") },		
    { LEN_STR_PAIR ("01;33") },		
    { 0, NULL },			
    { 0, NULL },			
    { LEN_STR_PAIR ("01;32") },		
    { LEN_STR_PAIR ("01;35") },		
    { LEN_STR_PAIR ("37;41") },		
    { LEN_STR_PAIR ("30;43") },		
    { LEN_STR_PAIR ("37;44") },		
    { LEN_STR_PAIR ("34;42") },		
    { LEN_STR_PAIR ("30;42") },		
    { LEN_STR_PAIR ("30;41") },		
    { 0, NULL },			
    { LEN_STR_PAIR ("\033[K") },	
  };
static struct color_ext_type *color_ext_list = NULL;
static enum Dereference_symlink dereference;
static enum
{
  
  IGNORE_MINIMAL
} ignore_mode;
struct ignore_pattern
  {
  };
static struct ignore_pattern *ignore_patterns,  *hide_patterns;
static struct quoting_options *filename_quoting_options;
enum { TIME_STAMP_LEN_MAXIMUM = MAX (1000, INT_STRLEN_BOUND (time_t)) };
enum
  {
    LS_FAILURE = 2
  };
enum
{
  AUTHOR_OPTION = CHAR_MAX + 1,
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
};
struct column_info
{
  bool valid_len;
  size_t line_len;
  size_t *col_arr;
};
static struct column_info *column_info;
#define MIN_COLUMN_WIDTH	3
#define DIRED_PUTCHAR(c) do {putchar ((c)); ++dired_pos;} while (0)
#define DIRED_FPUTS(s, stream, s_len) \
    do {fputs (s, stream); dired_pos += s_len;} while (0)
#define DIRED_FPUTS_LITERAL(s, stream) \
    do {fputs (s, stream); dired_pos += sizeof (s) - 1;} while (0)
#define DIRED_INDENT()							\
    do									\
      {									\
        if (dired)							\
          DIRED_FPUTS_LITERAL ("  ", stdout);				\
      }									\
    while (0)
#define PUSH_CURRENT_DIRED_POS(obs)					\
  do									\
    {									\
      if (dired)							\
        obstack_grow (obs, &dired_pos, sizeof (dired_pos));		\
    }									\
  while (0)
enum { MAX_MON_WIDTH = 5 };
static char abmon[12][MAX_MON_WIDTH * 2 * MB_LEN_MAX + 1];

int
main (int argc, char **argv)
;
enum parse_state
  {
    PS_FAIL
  };





static void
get_link_name (char const *filename, struct fileinfo *f, bool command_line_arg)
{
  f->linkname = areadlink_with_size (filename, f->stat.st_size);
  if (f->linkname == NULL)
    file_failure (command_line_arg, _("cannot read symbolic link %s"),
                  filename);
}
static char *
make_link_name (char const *name, char const *linkname)
;

static int
xstrcoll (char const *a, char const *b)
;
typedef void const *V;
typedef int (*qsortFunc)(V a, V b);
#define DIRFIRST_CHECK(a, b)						\
  do									\
    {									\
      bool a_is_dir = is_directory ((struct fileinfo const *) a);	\
      bool b_is_dir = is_directory ((struct fileinfo const *) b);	\
      if (a_is_dir && !b_is_dir)					\
        return -1;         			\
      if (!a_is_dir && b_is_dir)					\
        return 1;          			\
    }									\
  while (0)
#define DEFINE_SORT_FUNCTIONS(key_name, key_cmp_func)			\
  					\
  static int xstrcoll_##key_name (V a, V b)				\
  { return key_cmp_func (a, b, xstrcoll); }				\
  static int strcmp_##key_name (V a, V b)				\
  { return key_cmp_func (a, b, strcmp); }				\
                                                                        \
  					\
  static int rev_xstrcoll_##key_name (V a, V b)				\
  { return key_cmp_func (b, a, xstrcoll); }				\
  static int rev_strcmp_##key_name (V a, V b)				\
  { return key_cmp_func (b, a, strcmp); }				\
                                                                        \
  					\
  static int xstrcoll_df_##key_name (V a, V b)				\
  { DIRFIRST_CHECK (a, b); return key_cmp_func (a, b, xstrcoll); }	\
  static int strcmp_df_##key_name (V a, V b)				\
  { DIRFIRST_CHECK (a, b); return key_cmp_func (a, b, strcmp); }	\
                                                                        \
  					\
  static int rev_xstrcoll_df_##key_name (V a, V b)			\
  { DIRFIRST_CHECK (a, b); return key_cmp_func (b, a, xstrcoll); }	\
  static int rev_strcmp_df_##key_name (V a, V b)			\
  { DIRFIRST_CHECK (a, b); return key_cmp_func (b, a, strcmp); }
static int
cmp_ctime (struct fileinfo const *a, struct fileinfo const *b,
           int (*cmp) (char const *, char const *))
;
static int
cmp_mtime (struct fileinfo const *a, struct fileinfo const *b,
           int (*cmp) (char const *, char const *))
;
static int
cmp_atime (struct fileinfo const *a, struct fileinfo const *b,
           int (*cmp) (char const *, char const *))
;
static int
cmp_size (struct fileinfo const *a, struct fileinfo const *b,
          int (*cmp) (char const *, char const *))
;
static int
cmp_name (struct fileinfo const *a, struct fileinfo const *b,
          int (*cmp) (char const *, char const *))
;
static int
cmp_extension (struct fileinfo const *a, struct fileinfo const *b,
               int (*cmp) (char const *, char const *))
;
DEFINE_SORT_FUNCTIONS (ctime, cmp_ctime)
DEFINE_SORT_FUNCTIONS (mtime, cmp_mtime)
DEFINE_SORT_FUNCTIONS (atime, cmp_atime)
DEFINE_SORT_FUNCTIONS (size, cmp_size)
DEFINE_SORT_FUNCTIONS (name, cmp_name)
DEFINE_SORT_FUNCTIONS (extension, cmp_extension)




#define LIST_SORTFUNCTION_VARIANTS(key_name)                        \
  {                                                                 \
    {                                                               \
      { xstrcoll_##key_name, xstrcoll_df_##key_name },              \
      { rev_xstrcoll_##key_name, rev_xstrcoll_df_##key_name },      \
    },                                                              \
    {                                                               \
      { strcmp_##key_name, strcmp_df_##key_name },                  \
      { rev_strcmp_##key_name, rev_strcmp_df_##key_name },          \
    }                                                               \
  }


static size_t
quote_name (FILE *out, const char *name, struct quoting_options const *options,
            size_t *width)
{
  char smallbuf[BUFSIZ];
  size_t len = quotearg_buffer (smallbuf, sizeof smallbuf, name, -1, options), displayed_width IF_LINT ( = 0);
  char *buf;
  if (len < sizeof smallbuf)
    buf = smallbuf;
  else
    {
      buf = alloca (len + 1);
      quotearg_buffer (buf, len + 1, name, -1, options);
    }
  enum quoting_style qs = get_quoting_style (options);
  if (qmark_funny_chars
      && (qs == shell_quoting_style || qs == shell_always_quoting_style
          || qs == literal_quoting_style))
    {
      if (MB_CUR_MAX > 1)
        {
          char const *p = buf,  *plimit = buf + len;
          char *q = buf;
          displayed_width = 0;
          while (p < plimit)
            switch (*p)
              {
                case ' ': case '!': case '"': case '#': case '%':
                case '&': case '\'': case '(': case ')': case '*':
                case '+': case ',': case '-': case '.': case '/':
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                case ':': case ';': case '<': case '=': case '>':
                case '?':
                case 'A': case 'B': case 'C': case 'D': case 'E':
                case 'F': case 'G': case 'H': case 'I': case 'J':
                case 'K': case 'L': case 'M': case 'N': case 'O':
                case 'P': case 'Q': case 'R': case 'S': case 'T':
                case 'U': case 'V': case 'W': case 'X': case 'Y':
                case 'Z':
                case '[': case '\\': case ']': case '^': case '_':
                case 'a': case 'b': case 'c': case 'd': case 'e':
                case 'f': case 'g': case 'h': case 'i': case 'j':
                case 'k': case 'l': case 'm': case 'n': case 'o':
                case 'p': case 'q': case 'r': case 's': case 't':
                case 'u': case 'v': case 'w': case 'x': case 'y':
                case 'z': case '{': case '|': case '}': case '~':
                  *q++ = *p++;
                  displayed_width += 1;
                  break;
                default:
                  {
                    mbstate_t mbstate = { 0, };
                    do
                      {
                        wchar_t wc,  w;
                        size_t bytes;
                        bytes = mbrtowc (&wc, p, plimit - p, &mbstate);
                        if (bytes == (size_t) -1)
                          {
                            p++;
                            *q++ = '?';
                            displayed_width += 1;
                            break;
                          }
                        if (bytes == (size_t) -2)
                          {
                            p = plimit;
                            *q++ = '?';
                            displayed_width += 1;
                            break;
                          }
                        if (bytes == 0)
                          bytes = 1;
                        w = wcwidth (wc);
                        if (w >= 0)
                          {
                            for (; bytes > 0; --bytes)
                              *q++ = *p++;
                            displayed_width += w;
                          }
                        else
                          {
                            p += bytes;
                            *q++ = '?';
                            displayed_width += 1;
                          }
                      }
                    while (! mbsinit (&mbstate));
                  }
                  break;
              }
          len = q - buf;
        }
      else
        {
          char *p = buf;
          char const *plimit = buf + len;
          while (p < plimit)
            {
              if (! isprint (to_uchar (*p)))
                *p = '?';
              p++;
            }
          displayed_width = len;
        }
    }
  else if (width != NULL)
    {
      if (MB_CUR_MAX > 1)
        displayed_width = mbsnwidth (buf, len, 0);
      else
        {
          char const *p = buf,  *plimit = buf + len;
          displayed_width = 0;
          while (p < plimit)
            {
              if (isprint (to_uchar (*p)))
                displayed_width++;
              p++;
            }
        }
    }
  if (out != NULL)
    fwrite (buf, 1, len, out);
  if (width != NULL)
    *width = displayed_width;
  return len;
}
static void
prep_non_filename_text (void)
;
static size_t
print_file_name_and_frills (const struct fileinfo *f, size_t start_col)
;

static void
put_indicator (const struct bin_str *ind)
{
  if (! used_color)
    {
      used_color = true;
      prep_non_filename_text ();
    }
  fwrite (ind->string, ind->len, 1, stdout);
}
static size_t
length_of_file_name_and_frills (const struct fileinfo *f)
{
  size_t len = 0, name_width;
  char buf[MAX (LONGEST_HUMAN_READABLE + 1, INT_BUFSIZE_BOUND (uintmax_t))];
  if (print_inode)
    len += 1 + (format == with_commas
                ? strlen (umaxtostr (f->stat.st_ino, buf))
                : inode_number_width);
  if (print_block_size)
    len += 1 + (format == with_commas
                ? strlen (! f->stat_ok ? "?"
                          : human_readable (ST_NBLOCKS (f->stat), buf,
                                            human_output_opts, ST_NBLOCKSIZE,
                                            output_block_size))
                : block_size_width);
  if (print_scontext)
    len += 1 + (format == with_commas ? strlen (f->scontext) : scontext_width);
  quote_name (NULL, f->name, filename_quoting_options, &name_width);
  len += name_width;
  if (indicator_style != none)
    {
      char c = get_type_indicator (f->stat_ok, f->stat.st_mode, f->filetype);
      len += (c != 0);
    }
  return len;
}
static void
print_horizontal (void)
{
  size_t filesno, pos = 0, cols = calculate_columns (false);
  struct column_info const *line_fmt = &column_info[cols - 1];
  struct fileinfo const *f = sorted_file[0];
  size_t name_length = length_of_file_name_and_frills (f);
  size_t max_name_length = line_fmt->col_arr[0];
  print_file_name_and_frills (f, 0);
  for (filesno = 1; filesno < cwd_n_used; ++filesno)
    {
      size_t col = filesno % cols;
      if (col == 0)
        {
          putchar ('\n');
          pos = 0;
        }
      else
        {
          indent (pos + name_length, pos + max_name_length);
          pos += max_name_length;
        }
      f = sorted_file[filesno];
      print_file_name_and_frills (f, pos);
      name_length = length_of_file_name_and_frills (f);
      max_name_length = line_fmt->col_arr[col];
    }
  putchar ('\n');
}
static void
indent (size_t from, size_t to)
{
  while (from < to)
    {
      if (tabsize != 0 && to / tabsize > (from + 1) / tabsize)
        {
          putchar ('\t');
          from += tabsize - from % tabsize;
        }
      else
        {
          putchar (' ');
          from++;
        }
    }
}
static void
attach (char *dest, const char *dirname, const char *name)
{
  const char *dirnamep = dirname;
  if (dirname[0] != '.' || dirname[1] != 0)
    {
      while (*dirnamep)
        *dest++ = *dirnamep++;
      if (dirnamep > dirname && dirnamep[-1] != '/')
        *dest++ = '/';
    }
  while (*name)
    *dest++ = *name++;
  *dest = 0;
}
static void
init_column_info (void)
{
  size_t i, max_cols = MIN (max_idx, cwd_n_used);
  static size_t column_info_alloc;
  if (column_info_alloc < max_cols)
    {
      size_t new_column_info_alloc;
      size_t *p;
      if (max_cols < max_idx / 2)
        {
          column_info = xnrealloc (column_info, max_cols,
                                   2 * sizeof *column_info);
          new_column_info_alloc = 2 * max_cols;
        }
      else
        {
          column_info = xnrealloc (column_info, max_idx, sizeof *column_info);
          new_column_info_alloc = max_idx;
        }
      {
        size_t column_info_growth = new_column_info_alloc - column_info_alloc, s = column_info_alloc + 1 + new_column_info_alloc, t = s * column_info_growth;
        if (s < new_column_info_alloc || t / column_info_growth != s)
          xalloc_die ();
        p = xnmalloc (t / 2, sizeof *p);
      }
      for (i = column_info_alloc; i < new_column_info_alloc; i++)
        {
          column_info[i].col_arr = p;
          p += i + 1;
        }
      column_info_alloc = new_column_info_alloc;
    }
  for (i = 0; i < max_cols; ++i)
    {
      size_t j;
      column_info[i].valid_len = true;
      column_info[i].line_len = (i + 1) * MIN_COLUMN_WIDTH;
      for (j = 0; j <= i; ++j)
        column_info[i].col_arr[j] = MIN_COLUMN_WIDTH;
    }
}
static size_t
calculate_columns (bool by_columns)
{
  size_t filesno, cols, max_cols = MIN (max_idx, cwd_n_used);		
  init_column_info ();
  for (filesno = 0; filesno < cwd_n_used; ++filesno)
    {
      struct fileinfo const *f = sorted_file[filesno];
      size_t name_length = length_of_file_name_and_frills (f), i;
      for (i = 0; i < max_cols; ++i)
        {
          if (column_info[i].valid_len)
            {
              size_t idx = (by_columns
                            ? filesno / ((cwd_n_used + i) / (i + 1))
                            : filesno % (i + 1));
              size_t real_length = name_length + (idx == i ? 0 : 2);
              if (column_info[i].col_arr[idx] < real_length)
                {
                  column_info[i].line_len += (real_length
                                              - column_info[i].col_arr[idx]);
                  column_info[i].col_arr[idx] = real_length;
                  column_info[i].valid_len = (column_info[i].line_len
                                              < line_length);
                }
            }
        }
    }
  for (cols = max_cols; 1 < cols; --cols)
    {
      if (column_info[cols - 1].valid_len)
        break;
    }
  return cols;
}
